/* -----------------------------------------------------------------------------
*
*  npch.c
*
*  ANTz - realtime 3D data visualization tools for the real-world, based on NPE.
*
*  ANTz is hosted at http://openantz.com and NPE at http://neuralphysics.org
*
*  Written in 2010-2012 by Shane Saxon - contact@saxondigital.net
*
*  Please see main.c for a complete list of additional code contributors.
*
*  To the extent possible under law, the author(s) have dedicated all copyright 
*  and related and neighboring rights to this software to the public domain
*  worldwide. This software is distributed without any warranty.
*
*  Released under the CC0 license, which is GPL compatible.
*
*  You should have received a copy of the CC0 Public Domain Dedication along
*  with this software (license file named COPYING.txt). If not, see
*  http://creativecommons.org/publicdomain/zero/1.0/
*
* --------------------------------------------------------------------------- */

#include "npch.h"

#include "../npio.h"
#include "../npctrl.h"
#include "npfile.h"
#include "../os/npos.h"
#include <time.h>


int GetTrackOffset (char** trackScanner, char* propertyName);
void MapTrackOffset (int trackOffset, NPnodePtr node, char* propertyName, void* dataRef);
void MapTrackToAttribute (int trackOffset, float* attributeValuePtr, void* dataRef);
void CreateTracks (int numberTracks, void *dataRef);
void SaveHeader (int offset, char* header, void *dataRef);
int SaveHeaders (char** headers, void *dataRef);
bool SaveData (int numberTracks, int offset, float** trackFloatData, unsigned char** trackUcharData, char** buffer);
void DumpChannel (char* str, pNPch ch);

bool loopAtEndOfTrack = true;	//zz,  move to data->io.ch


//------------------------------------------------------------------------------
int GetTrackOffset(char** trackScanner, char* propertyName)
{
	int offset = 0;

	while (*trackScanner) 
	{
		if (strstr(*trackScanner, propertyName) != NULL)
			return offset;

		offset++;
		trackScanner++;
	}

	return -1;
}

//------------------------------------------------------------------------------
void MapTrackToFloatAttribute (int trackOffset, float* attributeValuePtr, void* dataRef)
{
	pData data = (pData) dataRef;

	*data->io.ch.nodePropertyFloatMemoryNextLocation++ = attributeValuePtr;
	data->io.ch.propertyFloatTracks[data->io.ch.propertyFloatTrackIndex++] = trackOffset;
}
//------------------------------------------------------------------------------
void MapTrackToUcharAttribute (int trackOffset, unsigned char* attributeValuePtr, void* dataRef)
{
	pData data = (pData) dataRef;

	*data->io.ch.nodePropertyUcharMemoryNextLocation++ = attributeValuePtr;
	data->io.ch.propertyUcharTracks[data->io.ch.propertyUcharTrackIndex++] = trackOffset;
}

//------------------------------------------------------------------------------
void MapTrackOffset(int trackOffset, NPnodePtr node, char* propertyName, void* dataRef)		//zz-JJ
{
	pData data = (pData) dataRef;

	if (strcmp(propertyName, "translate_x") == 0) MapTrackToFloatAttribute (trackOffset, &node->translate.x, data);
	if (strcmp(propertyName, "translate_y") == 0) MapTrackToFloatAttribute (trackOffset, &node->translate.y, data);
	if (strcmp(propertyName, "translate_z") == 0) MapTrackToFloatAttribute (trackOffset, &node->translate.z, data);
	if (strcmp(propertyName, "rotate_x") == 0) MapTrackToFloatAttribute (trackOffset, &node->rotate.x, data);
	if (strcmp(propertyName, "rotate_y") == 0) MapTrackToFloatAttribute (trackOffset, &node->rotate.y, data);
	if (strcmp(propertyName, "rotate_z") == 0) MapTrackToFloatAttribute (trackOffset, &node->rotate.z, data);

	printf("scale values currently x,y,z %f,%f, %f", node->scale.x, node->scale.y, node->scale.z);
	if (strcmp(propertyName, "scale_x") == 0) MapTrackToFloatAttribute (trackOffset, &node->scale.x, data);
	if (strcmp(propertyName, "scale_y") == 0) MapTrackToFloatAttribute (trackOffset, &node->scale.y, data);
	if (strcmp(propertyName, "scale_z") == 0) MapTrackToFloatAttribute (trackOffset, &node->scale.z, data);

	if (strcmp(propertyName, "translate_rate_x") == 0) MapTrackToFloatAttribute (trackOffset, &node->translateRate.x, data);
	if (strcmp(propertyName, "translate_rate_y") == 0) MapTrackToFloatAttribute (trackOffset, &node->translateRate.y, data);
	if (strcmp(propertyName, "translate_rate_z") == 0) MapTrackToFloatAttribute (trackOffset, &node->translateRate.z, data);
	if (strcmp(propertyName, "rotate_rate_x") == 0) MapTrackToFloatAttribute (trackOffset, &node->rotateRate.x, data);
	if (strcmp(propertyName, "rotate_rate_y") == 0) MapTrackToFloatAttribute (trackOffset, &node->rotateRate.y, data);
	if (strcmp(propertyName, "rotate_rate_z") == 0) MapTrackToFloatAttribute (trackOffset, &node->rotateRate.z, data);
	if (strcmp(propertyName, "scale_rate_x") == 0) MapTrackToFloatAttribute (trackOffset, &node->scaleRate.x, data);
	if (strcmp(propertyName, "scale_rate_y") == 0) MapTrackToFloatAttribute (trackOffset, &node->scaleRate.y, data);
	if (strcmp(propertyName, "scale_rate_z") == 0) MapTrackToFloatAttribute (trackOffset, &node->scaleRate.z, data);

//	if (strcmp(propertyName, "tag_mode") == 0) MapTrackToAttribute (trackOffset, &node->tagMode);
	if (strcmp(propertyName, "tag_offset_x") == 0) MapTrackToFloatAttribute (trackOffset, &node->tagOffset.x, data);
	if (strcmp(propertyName, "tag_offset_y") == 0) MapTrackToFloatAttribute (trackOffset, &node->tagOffset.y, data);
	if (strcmp(propertyName, "tag_offset_z") == 0) MapTrackToFloatAttribute (trackOffset, &node->tagOffset.z, data);

//	if (strcmp(propertyName, "geometry") == 0) MapTrackToAttribute (trackOffset, &node->geometry, data);
	if (strcmp(propertyName, "ratio") == 0) MapTrackToFloatAttribute (trackOffset, &node->ratio, data);
/*	if (strcmp(propertyName, "topo") == 0) MapTrackToAttribute (trackOffset, &node->topo, data);
	if (strcmp(propertyName, "facet") == 0) MapTrackToAttribute (trackOffset, &node->facet, data);

	if (strcmp(propertyName, "colorIndex") == 0) MapTrackToAttribute (trackOffset, &node->colorIndex, data);
*/	if (strcmp(propertyName, "color_r") == 0) MapTrackToUcharAttribute (trackOffset, &node->color.r, data);
	if (strcmp(propertyName, "color_g") == 0) MapTrackToUcharAttribute (trackOffset, &node->color.g, data);
	if (strcmp(propertyName, "color_b") == 0) MapTrackToUcharAttribute (trackOffset, &node->color.b, data);
	if (strcmp(propertyName, "color_a") == 0) MapTrackToUcharAttribute (trackOffset, &node->color.a, data); 

}		//zz-JJ-End

//------------------------------------------------------------------------------
void CreateTracks (int numberTracks, void *dataRef)
{
	pData data = (pData) dataRef;
	pNPch ch = &data->io.ch;
	int i = 0;
	float** trackFloatDataScanner = NULL;
	unsigned char** trackUcharDataScanner = NULL;
	char* trackBuffer = NULL;

	ch->trackFloatData = (float**) malloc (sizeof(float*) * (numberTracks+1));
	if (ch->trackFloatData == NULL)
	{
		npPostMsg ("err 2944 - malloc failed, CreateTracks", kNPmsgErr, data);
		return;
	}
	ch->trackUcharData = (unsigned char**) malloc (sizeof(unsigned char*) * (numberTracks+1));
	if (ch->trackUcharData == NULL)
	{
		npPostMsg ("err 2944 - malloc failed, CreateTracks", kNPmsgErr, data);
		return;
	}

	trackFloatDataScanner = ch->trackFloatData;
	trackUcharDataScanner = ch->trackUcharData;

	for (i = 0; i < numberTracks; i++) 
	{
		// allocate and initialize the float track buffers
		trackBuffer = (char*) malloc (sizeof(float*) * kNPtrackBufferSize);
		if (trackBuffer == NULL)
		{
			npPostMsg ("err 2945 - malloc failed, CreateTracks", kNPmsgErr, data);
			return;
		}
		memset(trackBuffer, 0, sizeof(float*) * kNPtrackBufferSize);
		*(trackFloatDataScanner++) = (float*)trackBuffer;
		*(trackFloatDataScanner) = NULL;

		// allocate and initialize the unsigned char track buffers
		trackBuffer = (char*) malloc (sizeof(unsigned char*) * kNPtrackBufferSize);
		if (trackBuffer == NULL)
		{
			npPostMsg ("err 2945 - malloc failed, CreateTracks", kNPmsgErr, data);
			return;
		}
		memset(trackBuffer, 0, sizeof(unsigned char*) * kNPtrackBufferSize);
		*(trackUcharDataScanner++) = (unsigned char*)trackBuffer;
		*(trackUcharDataScanner) = NULL;
	}
}

//------------------------------------------------------------------------------
void SaveHeader(int offset, char* header, void *dataRef)
{
	pData data = (pData) dataRef;
	char* headerCopy = NULL;
	
	headerCopy = (char*) malloc (strlen(header) + 1);
	if (headerCopy == NULL)
	{
		npPostMsg ("err 2946 - malloc failed, SaveHeader", kNPmsgErr, data);
		return;
	}

	strcpy (headerCopy, header);

	*(data->io.ch.trackNames + offset) = headerCopy;
	offset++;
	*(data->io.ch.trackNames + offset) = NULL;

	printf("Just saved header %s, now have %d headers\n", header, offset);
	// If we've reached limit, need to re-allocate both lists
}

//------------------------------------------------------------------------------
int SaveHeaders(char** headers, void *dataRef)
{
	// NULL terminated list of pointers
	int offset = 0;

	while (*headers != NULL)
		SaveHeader (offset++, *headers++, dataRef);

	return offset;
}

// saves data for each track into the different track data buffers
//------------------------------------------------------------------------------
bool SaveData(int numberTracks, int offset, float** trackData, unsigned char** trackUcharData, char** buffer)
{
	int i = 0;
	float* singleFloatTrack = NULL;
	unsigned char* singleUcharTrack = NULL;

	if (buffer == NULL) return false;

	for (i = 0; i < numberTracks; i++, buffer++) 
	{
		singleFloatTrack = *(trackData + i);
		*(singleFloatTrack + offset) = (float)atof(*buffer);
		singleUcharTrack = *(trackUcharData + i);
		*(singleUcharTrack + offset) = (unsigned char)atoi(*buffer);
	}

	return true;
}
/////////////////////////////////
// end of support functions
/////////////////////////////////


//init variables, allocate RAM
//------------------------------------------------------------------------------
void npInitCh (void* dataRef)
{
	pData data = (pData) dataRef;
	pNPch ch = &data->io.ch;
	int i = 0;

	ch->xrefIndex = 0;

	for (i = 0; i < kNPmaxTrackToAttributeMappings; i++)
	{
		ch->xref[i].attributeName = NULL;
		ch->xref[i].trackOffset = 0;
		ch->xref[i].channelNumber = 0;
	}

	ch->updateData = true;
	ch->channelReadIndex = kNPtrackBufferSize - 1;
	ch->channelWriteIndex = 0;
	ch->channelSampleRate = 60;	//zz-JJ needs to be a float
	ch->channelDataSize = 1;
				// wait a year before spontaneously starting, 1000*60*60*24*365;
	ch->channelNextDataChangeTime = clock() + 31536000000;
	ch->numberTracks = 0;
							// +1 for NULL pointer indicating end of list
	ch->trackNames = (char**) malloc(sizeof(char*) * (kNPmaxTracks+1));
	if (ch->trackNames == NULL)
		npPostMsg ("err 7657 - malloc failed, npInitCh", kNPmsgErr, data);

	ch->nodePropertyFloatMemoryNextLocation = &ch->nodePropertyFloatMemoryLocations[0];
	ch->nodePropertyUcharMemoryNextLocation = &ch->nodePropertyUcharMemoryLocations[0];
	ch->propertyFloatTrackIndex = 0;
	ch->propertyUcharTrackIndex = 0;
	ch->trackFloatData = NULL;
	ch->trackUcharData = NULL;

	for (i = 0; i < kNPmaxFloatProperties; i++)
	{
		ch->propertyFloatTracks[i] = 0;
		ch->nodePropertyFloatMemoryLocations[i] = NULL;
	}
	for (i = 0; i < kNPmaxUcharProperties; i++)
	{
		ch->propertyUcharTracks[i] = 0;
		ch->nodePropertyUcharMemoryLocations[i] = NULL;
	}
	ch->pause = false;
}

//------------------------------------------------------------------------------
void npCloseCh (void* dataRef)		//clean-up
{
	return;
}

//------------------------------------------------------------------------------
void DumpChannel(char* str, pNPch ch)
{
	printf("%s\n", str);
	printf(" channelNextDataChangeTime: %d\n", ch->channelNextDataChangeTime);
	printf(" channelReadIndex: %d\n", ch->channelReadIndex);
	printf(" channelSampleRate: %d\n", ch->channelSampleRate);
	printf(" channelWriteIndex: %d\n", ch->channelWriteIndex);
	//printf(" : %d\n", channel->nodePropertyMemoryLocations);
	//printf(" : %d\n", channel->nodePropertyMemoryNextLocation);
	printf(" numberTracks: %d\n", ch->numberTracks);
	printf(" propertyFloatTrackIndex: %d\n", ch->propertyFloatTrackIndex);
	//printf(" : %d\n", channel->propertyTracks);
	//printf(" : %d\n", channel->trackData);
	//printf(" : %d\n", channel->trackNames);
}


//  Currently have two options for behavior when we reach the end:
//  - pause
//  - loop
//
// If threads are supported this function will be called in its own thread.  
// In that case, it will implement circular buffer as part of loading data -- 
// waiting for data to be read before overwriting it.
// With looping with files larger than the circular buffer size, the large files
// are re-opened and re-read. This allows looping to occur with any size file.
//
// If threads are not supported, the buffer loading stops as soon as the buffer
// is full. The remaining data is never loaded.
// Looping occurs only on the data loaded into the buffer.
//------------------------------------------------------------------------------
void loadTracksIntoCircularBuffer (void* dataRef)
{
	pData data = (pData)dataRef;
	pNPch ch = &data->io.ch;
	FILE* fp = NULL;

	fp = ch->fileDataSource;

	while ( fp != NULL ) 
	{
		while (SaveData (ch->numberTracks, ch->channelWriteIndex++,
			ch->trackFloatData, ch->trackUcharData, npFileTokens(fp, data)) )
		{
			if (ch->channelWriteIndex > ch->channelDataSize) 
				ch->channelDataSize = ch->channelWriteIndex;

			// circular buffer requires thread, otherwise, we are totally done
			if (nposSupportsAntzThreads()) 
			{
				if (ch->channelWriteIndex == kNPtrackBufferSize)
					ch->channelWriteIndex = 0;
				//printf("channelReadIndex %d, channelWriteIndex %d, channelDataSize now %d\n", data->io.ch.channelReadIndex, data->io.ch.channelWriteIndex, data->io.ch.channelDataSize);
				
				 // wait for data to be read before overwriting
				while (ch->channelWriteIndex == ch->channelReadIndex) 
				{
					printf ("Waiting for data to be read..\n");
					nposSleep(0.1);
				}
			}		// without threads, break out
			else if (ch->channelWriteIndex == kNPtrackBufferSize)
				break;
		}
		npFileClose (fp, dataRef);
		fp = ch->fileDataSource = NULL;

		// if we are looping, and the buffer cycled, re-read the file
		if ( loopAtEndOfTrack 
			&& ch->channelDataSize == kNPtrackBufferSize
			&& nposSupportsAntzThreads() ) 
		{
			data->io.ch.channelReadIndex = kNPtrackBufferSize - 1;
			data->io.ch.channelWriteIndex = 0;
			fp = ch->fileDataSource = npFileOpen (ch->fullFilePath, "r", data);
		}
	}
	nposEndThread();
}

//open the file and read contents into a buffer
//determine the file type, and call either npChMapImport() or npChDataImport()
//------------------------------------------------------------------------------
//
// The channel is the data pipe between the tracks and the nodes
//
// Format is:
//
// id,track_set_id,track_column,attribute_name,id1,id2,id3
//
// For now, only using track_column, attribute_name
//
// track_column is the column offset in the data loaded with npLoadChTracks
//  
// data loaded here goes into xref[] array of NPxref that map a track offset to
// an attribute name
//------------------------------------------------------------------------------
void npLoadChMap (const char* filePath, void* dataRef)
{
	pData data = (pData) dataRef;
	FILE* file;
	char** tokens;
	char** tokenScanner;
	int tokenCount = 0;
	char path[2048];
	char* fullFilePath = NULL;
	int size;

	printf("npLoadChMap %s\n", filePath);
	npGetCWD (path, &size);
	fullFilePath = strcat (path, filePath);
	file = npFileOpen (fullFilePath, "r", data);

	if (file != NULL) 
	{
		tokens = tokenScanner = npFileTokens (file, data);

		while (*tokenScanner++) tokenCount++;		//count tokens
		if (tokenCount != 7) 	// expecting 7 columns, if not then err, abort
		{ 
			npPostMsg ("err 4588 - Invalid ANTzChMap file", kNPmsgErr, data);
			return;
		}

		// first line was header, now start with data
		tokens = npFileTokens(file, data);
		while (tokens != NULL) 
		{
			data->io.ch.xref[data->io.ch.xrefIndex].attributeName = 
					npNewStrcpy(*(tokens + kNPattributeNameOffset), data);
			data->io.ch.xref[data->io.ch.xrefIndex].trackOffset = 
					atoi(*(tokens + kNPtrackOffset));
			data->io.ch.xref[data->io.ch.xrefIndex].channelNumber = atoi(*(tokens + kNPchannelOffset));

			printf("Mapped track offset %d to attribute %s\n",
					data->io.ch.xref[data->io.ch.xrefIndex].trackOffset,
					data->io.ch.xref[data->io.ch.xrefIndex].attributeName);

			if (data->io.ch.xrefIndex < kNPmaxTrackToAttributeMappings)
				data->io.ch.xrefIndex++;

			tokens = npFileTokens (file, data);
		}
	}
}


// The track file is a text form of multiple tracks
// The first CSV line is column headers, free form text, not used by program,
// can be used to describe track
// Each column of data is a single track
//------------------------------------------------------------------------------
void npLoadChTracks (const char* filePath, void* dataRef)
{
	pData data = (pData) dataRef;
	pNPch ch = &data->io.ch;
	int size = 0;
	char path[2048];
	char* fullFilePath = NULL;
	FILE* file = NULL;
	float** trackData = NULL;

	static char npBuf[100];

	sprintf(npBuf, "npLoadChTracks %s\n", filePath);
	npPostMsg (npBuf, kNPmsgDebug, dataRef);

	npGetCWD (path, &size);
	fullFilePath = npNewStrcpy (strcat (path, filePath), data);
	file = npFileOpen (fullFilePath, "r", data);

	if (file != NULL)
	{
		npPostMsg ("..successfully opened file", kNPmsgDebug, dataRef);
		ch->fullFilePath = fullFilePath;
		ch->fileDataSource = file;
		npPostMsg ("..calling SaveHeaders", kNPmsgDebug, dataRef);
		ch->numberTracks = SaveHeaders (npFileTokens (file, data), data);
		npPostMsg ("..calling CreateTracks", kNPmsgDebug, dataRef);
		CreateTracks (ch->numberTracks, data);

		// data loaded in another thread to allow circular buffer, new thread is
		// for filling the buffer from the file this thread reads from circular
		// buffer, the new thread we are creating here writes to the circular
		// buffer both reading and writing threads follow boundary convention 
		// to implement circular buffer
		// 
		// circular buffer conventions:
		// writeIndex != readIndex means data can be written at the current index
		// writeIndex == readIndex means the circular buffer is full, nothing more
		//    can be written into it
		//  ((readIndex + 1)%circularBufferSize) == writeIndex means there is no
		//    data available to read
		//
		npPostMsg ("..calling nposBeginThread.. to actually load the data", kNPmsgDebug, dataRef);
		nposBeginThread (loadTracksIntoCircularBuffer, data);
		npPostMsg ("Done Loading Tracks", kNPmsgCtrl, dataRef);

	} 
	else {
		sprintf(npBuf, "Failed opening %s\n", fullFilePath);
		npPostMsg (npBuf, kNPmsgErr, dataRef);
	}
}

// load track metadata such as name, sample rate, data type, time-stamp, etc...
//------------------------------------------------------------------------------
void npLoadChMeta (const char* filePath, void* dataRef)
{
	return;
}

//passed in a buffer containing the entire contents of the antzchmap000x.csv
//parses the CSV formatted contents and builds the channel map list
//merging by using multiple imports allowed
//------------------------------------------------------------------------------
void npChMapImport (char* buffer, int size, void* dataRef)
{
	return;
}

//------------------------------------------------------------------------------
void npChDataImport (char* buffer, int size, void* dataRef)	//antzch0001.csv
{
	return;
}

//exports entire channel map to the buffer in CSV format
//the function that calls this is responsible for freeing the buffer when done
//typically would be written to a CSV file or DB then the buffer is freed
//------------------------------------------------------------------------------
void npChMapExport (char* buffer, int *size, void* dataRef)
{
	return;
}

//------------------------------------------------------------------------------
void npChDataExport (char* buffer, int *size, void* dataRef)
{
	return;
}

//all functions can be used to read/write to or from a CSV file or DB
//separate out the buffer to make this possible...
//add or remove ch attribute track pairs
//------------------------------------------------------------------------------
void npChMapAttributeAdd (int channel, int attribute, int track, void* dataRef)
{
	return;
}

//------------------------------------------------------------------------------
void npChMapAttributeDelete (int channel, int attribute, void* dataRef)
{
	return;
}

//------------------------------------------------------------------------------
void npNewChannel (int channel, void* dataRef)
{
	return;
}

//------------------------------------------------------------------------------
void npDeleteChannel (int channel, void* dataRef)
{
	return;
}

								//will remove '.x' from node->chInputID debug zz
//the node->channe.x specifies the channel map to use for attribute mapping
//of the track-attrib pairs added to the ptr-ptr list
//chID = 0 is the null channel, results in no channel mapping
//------------------------------------------------------------------------------
void npChSubscribeNode (NPnodePtr node, void* dataRef)
{
	pData data = (pData) dataRef;
	pNPch ch = &data->io.ch;
	int i = 0;
	static char msg[100];
	sprintf(msg, "..calling npChSubscribeNode, id %d", node->id);
	npPostMsg (msg, kNPmsgCtrl, dataRef);
	if (node->chInputID == 0)	{//zz
		npPostMsg ("..returning, chInputID is 0", kNPmsgCtrl, dataRef);
		return;
	}

	ch->channelNextDataChangeTime = clock();
	ch->channelReadIndex = 0;

	for (i = 0; i < ch->xrefIndex; i++) {
		sprintf(msg, "check xref %d", i);
		npPostMsg (msg, kNPmsgDebug, dataRef);
		if (node->chInputID == ch->xref[i].channelNumber) {
			sprintf(msg, "found channel, mapping attribute %s", ch->xref[i].attributeName);
			npPostMsg(msg, kNPmsgDebug, dataRef);
			MapTrackOffset(ch->xref[i].trackOffset, node, ch->xref[i].attributeName, data);
		}
	}

	printf("npChSubscribeNode, nodePropertyMemoryScanner starts at 0x%x, data->io.ch.nodePropertyMemoryNextLocation is 0%x\n",
	    &ch->nodePropertyFloatMemoryLocations[0],ch->nodePropertyFloatMemoryNextLocation);
}

//-----------------------------------------------------
// cycle backwards through the channel numbers present in the most recently
// loaded ChMap file.  The current channel number is a parameter, returns
// the next lower channel number (if there is one), or the greatest number channel
// number (if there is no channel number lower than the current one)
//
int npChGetChannelBefore(pNPch ch, int channelNumber) 
{
	int result = -1;  // -1 means "not found"
	int maximumChannelNumber = channelNumber;
	int i = 0;
	int channelNumberCandidate;

	for (i = 0; i < ch->xrefIndex; i++)
	{
		// if we have a "before" candidate
		channelNumberCandidate = ch->xref[i].channelNumber;
		if (channelNumberCandidate < channelNumber)
		{
			if (result == -1) // this is first candidate, save it
			{
				result = channelNumberCandidate;
			}
			else if ((channelNumber - result) > (channelNumber - channelNumberCandidate))
			{	// check if this candidate is closer to the channelNumber than the previous candidate
				// replace if result if this is the case
				result = channelNumberCandidate;
			}
		}
		// keep maximum up-to-date in case we cycle that way
		if (channelNumberCandidate > maximumChannelNumber)
		{
			//if ch = 0 then rollover to the last channel else set to 0
			if (channelNumber == 0)								//zz-JJ
				maximumChannelNumber = channelNumberCandidate;
			else
				maximumChannelNumber = 0;						//zz-JJ
		}
	}

	return (result == -1) ? maximumChannelNumber : result;
}

//-----------------------------------------------------
// cycle forward through the channel numbers present in the most recently
// loaded ChMap file.  The current channel number is a parameter, returns
// the next higher channel number (if there is one), or the smallest number channel
// number (if there is no channel number higher than the current one)
//
int npChGetChannelAfter(pNPch ch, int channelNumber) 
{
	int result = -1;  // -1 means "not found"
	int minimumChannelNumber = channelNumber;
	int i = 0;
	int channelNumberCandidate;

	printf("npChGetChannelAfter %d\n", channelNumber);
	for (i = 0; i < ch->xrefIndex; i++)
	{
		// if we have a "before" candidate
		channelNumberCandidate = ch->xref[i].channelNumber;
		if (channelNumberCandidate > channelNumber)
		{
			if (result == -1) 			// this is first candidate, save it
				result = channelNumberCandidate;
			else 
			{
				// check if this candidate is closer to the channelNumber than the previous candidate
				// replace if result if this is the case
				if ((result - channelNumber) > (channelNumberCandidate - channelNumber))
					result = channelNumberCandidate;
			}
		}
		// keep maximum up-to-date in case we cycle that way
		if (channelNumberCandidate < minimumChannelNumber)
			minimumChannelNumber = 0; //channelNumberCandidate; //zz-JJ
	}
	printf("..result is %d\n", (result == -1) ? minimumChannelNumber : result);

	return (result == -1) ? minimumChannelNumber : result;
}

//removes any pointers in the list associated with the node passed in
//used when changing channels which first removes the node, 
//then changes the chID and calls npChSubscribeNode to setup the new channel ptrs
//also called by npNodeDelete()
//------------------------------------------------------------------------------
void npChRemoveNode (NPnodePtr node, void* dataRef)
{
	pData data = (pData) dataRef;
	pNPch ch = &data->io.ch;
	char* nodeStart = (char*)node;
	char* nodeEnd = NULL;
	float **nodePropertyMemoryScanner =  NULL;

	nodeEnd = (nodeStart + sizeof(struct NPnode));
	nodePropertyMemoryScanner =  &ch->nodePropertyFloatMemoryLocations[0];

	while (nodePropertyMemoryScanner != ch->nodePropertyFloatMemoryNextLocation) 
	{
		if ( *nodePropertyMemoryScanner >= (float*)nodeStart
			&& *nodePropertyMemoryScanner < (float*)nodeEnd ) 
		{
			ch->nodePropertyFloatMemoryNextLocation--;
			*nodePropertyMemoryScanner =*(ch->nodePropertyFloatMemoryNextLocation);
			*(ch->nodePropertyFloatMemoryNextLocation) = NULL;
		}
		else
			nodePropertyMemoryScanner++;
	}
	printf("npChRemoveNode, nodePropertyMemoryScanner starts at 0x%x, data->io.ch.nodePropertyMemoryNextLocation is 0%x\n",
		&ch->nodePropertyFloatMemoryLocations[0], ch->nodePropertyFloatMemoryNextLocation);
}


//------------------------------------------------------------------------------
void npUpdateNodes(void *dataRef)
{
	pData data = (pData) dataRef;
	pNPch ch = &data->io.ch;
	float **nodePropertyFloatMemoryScanner = NULL;
	float *track = NULL;
	unsigned char **nodePropertyUcharMemoryScanner = NULL;
	unsigned char* ucharTrack = NULL;
	int channelScanIdx = 0;

	if (ch->updateData)
	{
		data->io.ch.updateData = false;

		// update all the float properties
		channelScanIdx = 0;
		nodePropertyFloatMemoryScanner = &ch->nodePropertyFloatMemoryLocations[0];
		while (nodePropertyFloatMemoryScanner != ch->nodePropertyFloatMemoryNextLocation)
		{
			track = *(ch->trackFloatData + ch->propertyFloatTracks[channelScanIdx]);
			**nodePropertyFloatMemoryScanner++ = *(track + ch->channelReadIndex);
			channelScanIdx++;
			if (ch->nodePropertyFloatMemoryNextLocation == &ch->nodePropertyFloatMemoryLocations[0]) break; // if reset while we are in loop
		}

		// update all the unsigned char properties (like r,g,b,a)
		channelScanIdx = 0;
		nodePropertyUcharMemoryScanner = &ch->nodePropertyUcharMemoryLocations[0];
		while (nodePropertyUcharMemoryScanner != ch->nodePropertyUcharMemoryNextLocation) 
		{
			ucharTrack = *(ch->trackUcharData + ch->propertyUcharTracks[channelScanIdx]);
			**nodePropertyUcharMemoryScanner++ = *(ucharTrack + ch->channelReadIndex);
			channelScanIdx++;
			if (ch->nodePropertyUcharMemoryNextLocation == &ch->nodePropertyUcharMemoryLocations[0]) break; // if reset while we are in loop
		}
	}
}

//(un)freeze a node in the list (or should we just add and del..) debug zz

//typically called once per cycle
//1st, update new input stream data, copy the last cycle into the track buffers
//2nd, update/calculate logical tracks, they may use raw tracks as inputs
//3rd, update nodes using track to node-attrib ptr lists
//lastly, output streams, such as live IO devices buffers or to a file
//------------------------------------------------------------------------------
void npUpdateCh (void* dataRef)
{
	pData data = (pData) dataRef;
	pNPch ch = &data->io.ch;
	clock_t now = 0;
	int nextReadIndex = 0;

	now = clock();

	if (data->io.ch.pause) return;  // nothing to update if we have paused
	if (now > ch->channelNextDataChangeTime) 
	{
		//printf("time for update... channelReadIndex %d, channelWriteIndex %d, channelDataSize %d\n", data->io.ch.channelReadIndex, data->io.ch.channelWriteIndex, data->io.ch.channelDataSize);
		ch->channelNextDataChangeTime += ch->channelSampleRate;
		nextReadIndex = (ch->channelReadIndex + 1) % (ch->channelDataSize);

		// if there is data to read, update the read index, and set the flag
		// that causes the updates
		if (nextReadIndex != ch->channelWriteIndex)
		{
			ch->updateData = true;
			ch->channelReadIndex = nextReadIndex;
//zz-JJ		printf("Reading %d, channelWriteIndex %d, channelDataSize %d\n", data->io.ch.channelReadIndex, data->io.ch.channelWriteIndex, data->io.ch.channelDataSize);
		}
	}
	npUpdateNodes(data);	//zz-JJ
}


//--- will add some stream functions
//multiple types DB, CSV, Network UDP & IP, LiveIO - audio, video, TNG3B, mouse
//updated with npUpdateCh() 
//------------------------------------------------------------------------------
void npChOpenInputStream (int needToFigureOut, void* dataRef)
{
	return;
}

//------------------------------------------------------------------------------
void npChCloseInputStream (int needToFigureOut, void* dataRef)
{
	return;
}

//------------------------------------------------------------------------------
void npChOpenOutputStream (int needToFigureOut, void* dataRef)
{
	return;
}

//------------------------------------------------------------------------------
void npChCloseOutStream (int needToFigureOut, void* dataRef)
{
	return;
}

//------------------------------------------------------------------------------
void npchPause (void* dataRef)	//zz-JJ
{
	pData data = (pData) dataRef;

	//toggle the pause state
	data->io.ch.pause = 1 - data->io.ch.pause;

	// reset next time to change, needed when resuming
	data->io.ch.channelNextDataChangeTime = clock();
}


//zz-JJ
//------------------------------------------------------------------------------
void npSyncChMap (void* dataRef)
{
	//delete entire channel pointer map list
	pData data = (pData) dataRef;
	data->io.ch.nodePropertyFloatMemoryNextLocation = &data->io.ch.nodePropertyFloatMemoryLocations[0];
	data->io.ch.nodePropertyUcharMemoryNextLocation = &data->io.ch.nodePropertyUcharMemoryLocations[0];

	//add all nodes with channels back to the pointer map list
	npTraverseMap (npChSubscribeNode, dataRef);
}

